/*
 * Licensed to Crate.io GmbH ("Crate") under one or more contributor
 * license agreements.  See the NOTICE file distributed with this work for
 * additional information regarding copyright ownership.  Crate licenses
 * this file to you under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.  You may
 * obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
 * License for the specific language governing permissions and limitations
 * under the License.
 *
 * However, if you have executed another commercial license agreement
 * with Crate these terms will supersede the license and you may use the
 * software solely pursuant to the terms of the relevant commercial agreement.
 */

package io.crate.analyze;

import static io.crate.testing.Asserts.exactlyInstanceOf;
import static io.crate.testing.Asserts.isAlias;
import static io.crate.testing.Asserts.isFunction;
import static io.crate.testing.Asserts.isLiteral;
import static io.crate.testing.Asserts.isReference;
import static io.crate.testing.Asserts.toCondition;
import static java.util.Collections.singletonList;
import static org.assertj.core.api.Assertions.assertThatThrownBy;
import static org.hamcrest.Matchers.instanceOf;
import static org.hamcrest.Matchers.is;
import static org.hamcrest.Matchers.not;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertThat;
import static org.junit.Assert.assertTrue;
import static org.mockito.Mockito.mock;

import java.io.IOException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

import io.crate.analyze.relations.DocTableRelation;
import io.crate.data.Row;
import io.crate.data.Row1;
import io.crate.data.RowN;
import io.crate.exceptions.ColumnUnknownException;
import io.crate.exceptions.ColumnValidationException;
import io.crate.exceptions.ConversionException;
import io.crate.exceptions.OperationOnInaccessibleRelationException;
import io.crate.exceptions.RelationUnknown;
import io.crate.exceptions.VersioningValidationException;
import io.crate.expression.operator.EqOperator;
import io.crate.expression.predicate.NotPredicate;
import io.crate.expression.symbol.Assignments;
import io.crate.expression.symbol.DynamicReference;
import io.crate.expression.symbol.Literal;
import io.crate.expression.symbol.ParameterSymbol;
import io.crate.expression.symbol.Symbol;
import io.crate.metadata.ColumnIdent;
import io.crate.metadata.Reference;
import io.crate.metadata.RelationName;
import io.crate.metadata.Schemas;
import io.crate.metadata.doc.DocTableInfo;
import io.crate.planner.DependencyCarrier;
import io.crate.planner.Plan;
import io.crate.planner.operators.SubQueryResults;
import io.crate.test.integration.CrateDummyClusterServiceUnitTest;
import io.crate.testing.Asserts;
import io.crate.testing.SQLExecutor;
import io.crate.testing.TestingRowConsumer;
import io.crate.types.ArrayType;
import io.crate.types.DataType;
import io.crate.types.DataTypes;
import io.crate.types.DoubleType;
import io.crate.types.ObjectType;

public class UpdateAnalyzerTest extends CrateDummyClusterServiceUnitTest {

    private SQLExecutor e;

    @Before
    public void prepare() throws IOException {
        SQLExecutor.Builder builder = SQLExecutor.builder(clusterService)
            .addTable(TableDefinitions.USER_TABLE_DEFINITION)
            .addTable(TableDefinitions.USER_TABLE_CLUSTERED_BY_ONLY_DEFINITION)
            .addPartitionedTable(
                TableDefinitions.TEST_PARTITIONED_TABLE_DEFINITION,
                TableDefinitions.TEST_PARTITIONED_TABLE_PARTITIONS)
            .addTable(
                "create table doc.nestedclustered (" +
                "   obj object as (" +
                "       name string" +
                "   )," +
                "   other_obj object" +
                ") clustered by (obj['name']) "
            )
            .addTable(
                "create table doc.t_nested_pk (" +
                "   o object as (" +
                "       x integer primary key," +
                "       y integer" +
                "   )" +
                ")"
            )
            .addTable("create table bag (id short primary key, ob array(object))")
            .addPartitionedTable(
                "create table doc.parted_generated_column (" +
                "   ts timestamp with time zone," +
                "   day as date_trunc('day', ts)" +
                ") partitioned by (day) "
            )
            .addPartitionedTable(
                "create table doc.nested_parted_generated_column (" +
                "   \"user\" object as (" +
                "       name string" +
                "   )," +
                "   name as concat(\"user\"['name'], 'bar')" +
                ") partitioned by (name) "
            );

        e = builder.build();
    }

    protected AnalyzedUpdateStatement analyze(String statement) {
        return e.analyze(statement);
    }

    @Test
    public void testUpdateAnalysis() throws Exception {
        AnalyzedStatement analyzedStatement = analyze("update users set name='Ford Prefect'");
        assertThat(analyzedStatement, instanceOf(AnalyzedUpdateStatement.class));
    }

    @Test(expected = RelationUnknown.class)
    public void testUpdateUnknownTable() throws Exception {
        analyze("update unknown set name='Prosser'");
    }

    @Test
    public void testUpdateSetColumnToColumnValue() throws Exception {
        AnalyzedUpdateStatement update = analyze("update users set name=name");
        assertThat(update.assignmentByTargetCol().size(), is(1));
        Symbol value = update.assignmentByTargetCol().entrySet().iterator().next().getValue();
        Asserts.assertThat(value).isReference("name");
    }

    @Test
    public void testUpdateSetExpression() throws Exception {
        AnalyzedUpdateStatement update = analyze("update users set other_id=other_id+1");
        assertThat(update.assignmentByTargetCol().size(), is(1));
        Symbol value = update.assignmentByTargetCol().entrySet().iterator().next().getValue();
        Asserts.assertThat(value).isFunction("add");
    }

    @Test(expected = IllegalArgumentException.class)
    public void testUpdateSameReferenceRepeated() throws Exception {
        analyze("update users set name='Trillian', name='Ford'");
    }

    @Test(expected = IllegalArgumentException.class)
    public void testUpdateSameNestedReferenceRepeated() throws Exception {
        analyze("update users set details['arms']=3, details['arms']=5");
    }

    @Test
    public void testUpdateSysTables() throws Exception {
        expectedException.expect(OperationOnInaccessibleRelationException.class);
        expectedException.expectMessage("The relation \"sys.nodes\" doesn't support or allow UPDATE " +
                                        "operations, as it is read-only.");
        analyze("update sys.nodes set fs={\"free\"=0}");
    }

    @Test
    public void testNumericTypeOutOfRange() {
        expectedException.expect(ColumnValidationException.class);
        expectedException.expectMessage("Validation failed for shorts: Cannot cast expression `-100000` of type `integer` to `smallint`");
        analyze("update users set shorts=-100000");
    }

    @Test
    public void testNumericOutOfRangeFromFunction() {
        expectedException.expect(ColumnValidationException.class);
        expectedException.expectMessage("Validation failed for bytes: Cannot cast expression `1234` of type `integer` to `byte`");
        analyze("update users set bytes=abs(-1234)");
    }

    @Test
    public void testUpdateAssignments() throws Exception {
        AnalyzedUpdateStatement update = analyze("update users set name='Trillian'");
        assertThat(update.assignmentByTargetCol().size(), is(1));
        assertThat(((DocTableRelation) update.table()).tableInfo().ident(), is(new RelationName(Schemas.DOC_SCHEMA_NAME, "users")));

        Reference ref = update.assignmentByTargetCol().keySet().iterator().next();
        assertThat(ref.ident().tableIdent().name(), is("users"));
        assertThat(ref.column().name(), is("name"));
        assertTrue(update.assignmentByTargetCol().containsKey(ref));

        Symbol value = update.assignmentByTargetCol().entrySet().iterator().next().getValue();
        Asserts.assertThat(value).isLiteral("Trillian");
    }

    @Test
    public void testUpdateAssignmentNestedDynamicColumn() throws Exception {
        AnalyzedUpdateStatement update = analyze("update users set details['arms']=3");
        assertThat(update.assignmentByTargetCol().size(), is(1));

        Reference ref = update.assignmentByTargetCol().keySet().iterator().next();
        assertThat(ref, instanceOf(DynamicReference.class));
        Assert.assertEquals(DataTypes.INTEGER, ref.valueType());
        assertThat(ref.column().isTopLevel(), is(false));
        assertThat(ref.column().fqn(), is("details.arms"));
    }

    @Test
    public void testUpdateAssignmentWrongType() throws Exception {
        expectedException.expect(ColumnValidationException.class);
        analyze("update users set other_id='String'");
    }

    @Test
    public void testUpdateAssignmentConvertableType() throws Exception {
        AnalyzedUpdateStatement update = analyze("update users set other_id=9.9");
        Reference ref = update.assignmentByTargetCol().keySet().iterator().next();
        assertThat(ref, not(instanceOf(DynamicReference.class)));
        assertEquals(DataTypes.LONG, ref.valueType());

        Assignments assignments = Assignments.convert(update.assignmentByTargetCol(), e.nodeCtx);
        Symbol[] sources = assignments.bindSources(
            ((DocTableInfo) update.table().tableInfo()), Row.EMPTY, SubQueryResults.EMPTY);
        Asserts.assertThat(sources[0]).isLiteral(9L);
    }

    @Test
    public void testUpdateMuchAssignments() throws Exception {
        AnalyzedUpdateStatement update = analyze(
            "update users set other_id=9.9, name='Trillian', details={}, stuff=true, foo='bar'");
        assertThat(update.assignmentByTargetCol().size(), is(5));
    }

    @Test
    public void testNoWhereClause() throws Exception {
        AnalyzedUpdateStatement update = analyze("update users set other_id=9");
        Asserts.assertThat(update.query()).isLiteral(true);
    }

    @Test
    public void testNoMatchWhereClause() throws Exception {
        AnalyzedUpdateStatement update = analyze("update users set other_id=9 where true=false");
        Asserts.assertThat(update.query()).isLiteral(false);
    }

    @Test
    public void testUpdateWhereClause() throws Exception {
        AnalyzedUpdateStatement update = analyze("update users set other_id=9 where name='Trillian'");
        Asserts.assertThat(update.query()).isFunction(EqOperator.NAME, isReference("name"), isLiteral("Trillian"));
    }

    @Test
    public void testQualifiedNameReference() throws Exception {
        expectedException.expect(IllegalArgumentException.class);
        expectedException.expectMessage("Column reference \"users.name\" has too many parts. A column must not have a schema or a table here.");
        analyze("update users set users.name='Trillian'");
    }

    @Test
    public void testUpdateWithParameter() throws Exception {
        AnalyzedUpdateStatement update = analyze("update users set name=?, other_id=?, friends=? where id=?");

        RelationName usersRelation = new RelationName("doc", "users");
        assertThat(update.assignmentByTargetCol().size(), is(3));
        DocTableInfo tableInfo = e.schemas().getTableInfo(usersRelation);
        Reference name = tableInfo.getReference(new ColumnIdent("name"));
        Reference friendsRef = tableInfo.getReference(new ColumnIdent("friends"));
        Reference otherId = tableInfo.getReference(new ColumnIdent("other_id"));
        assertThat(update.assignmentByTargetCol().get(name), instanceOf(ParameterSymbol.class));
        assertThat(update.assignmentByTargetCol().get(friendsRef), instanceOf(ParameterSymbol.class));
        assertThat(update.assignmentByTargetCol().get(otherId), instanceOf(ParameterSymbol.class));

        Asserts.assertThat(update.query())
            .isFunction(EqOperator.NAME, isReference("id"), exactlyInstanceOf(ParameterSymbol.class));
    }

    @Test
    public void testUpdateWithWrongParameters() throws Exception {
        Object[] params = {
            List.of(new HashMap<String, Object>()),
            new Map[0],
            new Long[]{1L, 2L, 3L}};
        AnalyzedUpdateStatement update = analyze("update users set name=?, friends=? where other_id=?");

        Assignments assignments = Assignments.convert(update.assignmentByTargetCol(), e.nodeCtx);
        expectedException.expect(ConversionException.class);
        expectedException.expectMessage("Cannot cast value `[{}]` to type `text`");
        assignments.bindSources(((DocTableInfo) update.table().tableInfo()), new RowN(params), SubQueryResults.EMPTY);
    }

    @Test
    public void testUpdateWithEmptyObjectArray() throws Exception {
        Object[] params = {new Map[0], 0};
        AnalyzedUpdateStatement update = analyze("update users set friends=? where other_id=0");

        Assignments assignments = Assignments.convert(update.assignmentByTargetCol(), e.nodeCtx);
        Symbol[] sources = assignments.bindSources(((DocTableInfo) update.table().tableInfo()), new RowN(params), SubQueryResults.EMPTY);


        assertThat(sources[0].valueType().id(), is(ArrayType.ID));
        assertThat(((ArrayType) sources[0].valueType()).innerType().id(), is(ObjectType.ID));
        assertThat(((List) ((Literal) sources[0]).value()).size(), is(0));
    }

    @Test
    public void testUpdateSystemColumn() throws Exception {
        expectedException.expect(ColumnValidationException.class);
        expectedException.expectMessage("Validation failed for _id: Updating a system column is not supported");
        analyze("update users set _id=1");
    }

    @Test
    public void testUpdatePrimaryKey() throws Exception {
        expectedException.expect(ColumnValidationException.class);
        analyze("update users set id=1");
    }

    @Test
    public void testUpdateClusteredBy() throws Exception {
        expectedException.expect(ColumnValidationException.class);
        expectedException.expectMessage("Validation failed for id: Updating a clustered-by column is not supported");
        analyze("update users_clustered_by_only set id=1");
    }

    @Test(expected = ColumnValidationException.class)
    public void testUpdatePartitionedByColumn() throws Exception {
        analyze("update parted set date = 1395874800000");
    }

    @Test
    public void testUpdatePrimaryKeyIfNestedDoesNotWork() throws Exception {
        expectedException.expect(ColumnValidationException.class);
        analyze("update t_nested_pk set o = {y=10}");
    }

    @Test
    public void testUpdateColumnReferencedInGeneratedPartitionByColumn() throws Exception {
        expectedException.expect(ColumnValidationException.class);
        expectedException.expectMessage("Updating a column which is referenced in a partitioned by generated column expression is not supported");
        analyze("update parted_generated_column set ts = 1449999900000");
    }

    @Test
    public void testUpdateColumnReferencedInGeneratedPartitionByColumnNestedParent() throws Exception {
        expectedException.expect(ColumnValidationException.class);
        expectedException.expectMessage("Updating a column which is referenced in a partitioned by generated column expression is not supported");
        analyze("update nested_parted_generated_column set \"user\" = {name = 'Ford'}");
    }

    @Test
    public void testUpdateTableAlias() throws Exception {
        AnalyzedUpdateStatement expected = analyze("update users set awesome=true where awesome=false");
        AnalyzedUpdateStatement actual = analyze("update users as u set awesome=true where awesome=false");

        assertThat(expected.assignmentByTargetCol(), is(actual.assignmentByTargetCol()));
        assertThat(expected.query(), is(actual.query()));
    }

    @Test
    public void testUpdateObjectArrayField() throws Exception {
        expectedException.expect(IllegalArgumentException.class);
        analyze("update users set friends['id'] = ?");
    }

    @Test
    public void testUpdateArrayByElement() {
        expectedException.expect(IllegalArgumentException.class);
        expectedException.expectMessage("Updating a single element of an array is not supported");
        analyze("update users set friends[1] = 2");
    }

    @Test
    public void test_update_array_of_objects_by_elements() {
        assertThatThrownBy(
            () -> analyze("update users set friends[1]['val1']['val2'] = true"))
            .isInstanceOf(IllegalArgumentException.class)
            .hasMessage("Updating a single element of an array is not supported");
    }

    @Test
    public void testWhereClauseObjectArrayField() throws Exception {
        expectedException.expect(UnsupportedOperationException.class);
        expectedException.expectMessage("Unknown function: (doc.users.friends['id'] = 5)," +
                                        " no overload found for matching argument types: (bigint_array, integer).");
        analyze("update users set awesome=true where friends['id'] = 5");
    }

    @Test
    public void testUpdateWithFQName() throws Exception {
        expectedException.expect(IllegalArgumentException.class);
        expectedException.expectMessage("Column reference \"users.name\" has too many parts. A column must not have a schema or a table here.");
        analyze("update users set users.name = 'Ford Mustang'");
    }

    @Test
    public void testUpdateDynamicNestedArrayParamLiteral() throws Exception {
        AnalyzedUpdateStatement update = analyze("update users set new=[[1.9, 4.8], [9.7, 12.7]]");
        DataType dataType = update.assignmentByTargetCol().values().iterator().next().valueType();
        assertThat(dataType, is(new ArrayType(new ArrayType(DoubleType.INSTANCE))));
    }

    @Test
    public void testUpdateDynamicNestedArrayParam() throws Exception {
        Object[] params = {
            new Object[]{
                new Object[]{
                    1.9, 4.8
                },
                new Object[]{
                    9.7, 12.7
                }
            }
        };
        AnalyzedUpdateStatement update = analyze("update users set new=? where id=1");
        Assignments assignments = Assignments.convert(update.assignmentByTargetCol(), e.nodeCtx);
        Symbol[] sources = assignments.bindSources(
            ((DocTableInfo) update.table().tableInfo()), new RowN(params), SubQueryResults.EMPTY);

        DataType dataType = sources[0].valueType();
        assertThat(dataType, is(new ArrayType(new ArrayType(DoubleType.INSTANCE))));
    }

    @Test
    public void testUpdateInvalidType() throws Exception {
        Object[] params = {
            new Object[]{
                new Object[]{"a", "b"}
            }
        };
        AnalyzedUpdateStatement update = analyze("update users set tags=? where id=1");

        expectedException.expect(ConversionException.class);
        expectedException.expectMessage("Cannot cast value `[[a, b]]` to type `text_array`");
        Assignments assignments = Assignments.convert(update.assignmentByTargetCol(), e.nodeCtx);
        assignments.bindSources(((DocTableInfo) update.table().tableInfo()), new RowN(params), SubQueryResults.EMPTY);
    }

    @Test
    public void testUsingFQColumnNameShouldBePossibleInWhereClause() throws Exception {
        AnalyzedUpdateStatement update = analyze("update users set name = 'foo' where users.name != 'foo'");
        Asserts.assertThat(update.query())
            .isFunction(NotPredicate.NAME, isFunction(EqOperator.NAME, isReference("name"), isLiteral("foo")));
    }

    @Test
    public void testTestUpdateOnTableWithAliasAndFQColumnNameInWhereClause() throws Exception {
        AnalyzedUpdateStatement update = analyze("update users  t set name = 'foo' where t.name != 'foo'");
        Asserts.assertThat(update.query())
            .isFunction(NotPredicate.NAME, isFunction(EqOperator.NAME, isReference("name"), isLiteral("foo")));
    }

    @Test
    public void testUpdateNestedClusteredByColumn() throws Exception {
        expectedException.expect(ColumnValidationException.class);
        expectedException.expectMessage("Validation failed for obj: Updating a clustered-by column is not supported");
        analyze("update nestedclustered set obj = {name='foobar'}");
    }

    @Test
    public void testUpdateNestedClusteredByColumnWithOtherObject() throws Exception {
        expectedException.expect(ColumnValidationException.class);
        expectedException.expectMessage("Validation failed for obj: Updating a clustered-by column is not supported");
        analyze("update nestedclustered set obj = other_obj");
    }

    @Test
    public void testUpdateWhereVersionUsingWrongOperator() throws Exception {
        expectedException.expect(VersioningValidationException.class);
        expectedException.expectMessage(VersioningValidationException.VERSION_COLUMN_USAGE_MSG);
        execute(e.plan("update users set text = ? where text = ? and \"_version\" >= ?"),
            new RowN(new Object[]{"already in panic", "don't panic", 3}));
    }

    @Test
    public void testUpdateWhereVersionIsColumn() throws Exception {
        expectedException.expect(VersioningValidationException.class);
        expectedException.expectMessage(VersioningValidationException.VERSION_COLUMN_USAGE_MSG);
        execute(e.plan("update users set col2 = ? where _version = id"), new Row1(1));
    }

    @Test
    public void testUpdateWhereVersionInOperatorColumn() throws Exception {
        expectedException.expect(VersioningValidationException.class);
        expectedException.expectMessage(VersioningValidationException.VERSION_COLUMN_USAGE_MSG);
        execute(e.plan("update users set col2 = 'x' where _version in (1,2,3)"), Row.EMPTY);
    }

    @Test
    public void testUpdateWhereVersionOrOperatorColumn() throws Exception {
        expectedException.expect(VersioningValidationException.class);
        expectedException.expectMessage(VersioningValidationException.VERSION_COLUMN_USAGE_MSG);
        execute(e.plan("update users set col2 = ? where _version = 1 or _version = 2"), new Row1(1));
    }


    @Test
    public void testUpdateWhereVersionAddition() throws Exception {
        expectedException.expect(VersioningValidationException.class);
        expectedException.expectMessage(VersioningValidationException.VERSION_COLUMN_USAGE_MSG);
        execute(e.plan("update users set col2 = ? where _version + 1 = 2"), new Row1(1));
    }

    @Test
    public void testUpdateWhereVersionNotPredicate() throws Exception {
        expectedException.expect(VersioningValidationException.class);
        expectedException.expectMessage(VersioningValidationException.VERSION_COLUMN_USAGE_MSG);
        execute(e.plan("update users set text = ? where not (_version = 1 and id = 1)"), new Row1(1));
    }

    @Test
    public void testUpdateWhereVersionOrOperator() throws Exception {
        expectedException.expect(VersioningValidationException.class);
        expectedException.expectMessage(VersioningValidationException.VERSION_COLUMN_USAGE_MSG);
        execute(e.plan("update users set awesome = true where _version = 1 or _version = 2"), Row.EMPTY);
    }

    @Test
    public void testUpdateWithVersionZero() throws Exception {
        expectedException.expect(VersioningValidationException.class);
        expectedException.expectMessage(VersioningValidationException.VERSION_COLUMN_USAGE_MSG);
        execute(e.plan("update users set awesome=true where name='Ford' and _version=0"), Row.EMPTY);
    }

    @Test
    public void testSelectWhereVersionIsNullPredicate() throws Exception {
        expectedException.expect(VersioningValidationException.class);
        expectedException.expectMessage(VersioningValidationException.VERSION_COLUMN_USAGE_MSG);
        execute(e.plan("update users set col2 = 'x' where _version is null"), Row.EMPTY);
    }

    @Test
    public void testUpdateElementOfObjectArrayUsingParameterExpressionResultsInCorrectlyTypedParameterSymbol() {
        AnalyzedUpdateStatement stmt = e.analyze("UPDATE bag SET ob = [?] WHERE id = ?");
        Asserts.assertThat(stmt.assignmentByTargetCol()).hasEntrySatisfying(
            toCondition(isReference("ob", new ArrayType<>(DataTypes.UNTYPED_OBJECT))),
            toCondition(isFunction("_array")));
        Asserts.assertThat(stmt.assignmentByTargetCol().values())
                .satisfiesExactly(l -> Asserts.assertThat(l)
                    .isFunction("_array", singletonList(DataTypes.UNTYPED_OBJECT)));
    }

    @Test
    public void testUpdateElementOfObjectArrayUsingParameterExpressionInsideFunctionResultsInCorrectlyTypedParameterSymbol() {
        AnalyzedUpdateStatement stmt = e.analyze("UPDATE bag SET ob = array_cat([?], [{obb=1}]) WHERE id = ?");
        Asserts.assertThat(stmt.assignmentByTargetCol()).hasEntrySatisfying(
            toCondition(isReference("ob", new ArrayType<>(DataTypes.UNTYPED_OBJECT))),
            toCondition(isFunction("array_cat")));
        Asserts.assertThat(stmt.assignmentByTargetCol().values()).satisfiesExactly(
            isFunction("array_cat",
                isFunction("_array", singletonList(DataTypes.UNTYPED_OBJECT)),
                       exactlyInstanceOf(Literal.class)));
    }

    @Test
    public void test_update_returning_with_asterisk_contains_all_columns_in_returning_clause() {
        AnalyzedUpdateStatement stmt = e.analyze(
            "UPDATE users SET name='noam' RETURNING *");
        Asserts.assertThat(stmt.assignmentByTargetCol()).hasEntrySatisfying(
            toCondition(isReference("name", DataTypes.STRING)),
            toCondition(isLiteral("noam")));
        assertThat(stmt.outputs().size(), is(17));
    }

    @Test
    public void test_update_returning_with_single_value_in_returning_clause() {
        AnalyzedUpdateStatement stmt = e.analyze(
            "UPDATE users SET name='noam' RETURNING id AS foo");
        Asserts.assertThat(stmt.assignmentByTargetCol()).hasEntrySatisfying(
            toCondition(isReference("name", DataTypes.STRING)),
            toCondition(isLiteral("noam")));
        Asserts.assertThat(stmt.outputs()).satisfiesExactly(isAlias("foo", isReference("id")));
    }

    @Test
    public void test_update_returning_with_multiple_values_in_returning_clause() {
        AnalyzedUpdateStatement stmt = e.analyze(
            "UPDATE users SET name='noam' RETURNING id AS foo, name AS bar");
        Asserts.assertThat(stmt.assignmentByTargetCol()).hasEntrySatisfying(
            toCondition(isReference("name", DataTypes.STRING)),
            toCondition(isLiteral("noam")));
        Asserts.assertThat(stmt.outputs()).satisfiesExactly(
            isAlias("foo", isReference("id")), isAlias("bar", isReference("name")));
    }

    @Test
    public void test_updat_returning_with_invalid_column_returning_error() {
        expectedException.expect(ColumnUnknownException.class);
        expectedException.expectMessage("Column invalid unknown");
        e.analyze("UPDATE users SET name='noam' RETURNING invalid");
    }

    @Test
    public void test_update_returning_with_single_value_altered_in_returning_clause() {
        AnalyzedUpdateStatement stmt = e.analyze(
            "UPDATE users SET name='noam' RETURNING id + 1 AS foo");
        Asserts.assertThat(stmt.assignmentByTargetCol()).hasEntrySatisfying(
            toCondition(isReference("name", DataTypes.STRING)),
            toCondition(isLiteral("noam")));
        Asserts.assertThat(stmt.outputs()).satisfiesExactly(
            isAlias("foo", isFunction("add", isReference("id"), isLiteral(1L))));
    }

    @Test
    public void test_update_returning_with_multiple_values_altered_in_returning_clause() {
        AnalyzedUpdateStatement stmt = e.analyze(
            "UPDATE users SET name='noam' RETURNING id + 1 AS foo, id -1 as bar");
        Asserts.assertThat(stmt.assignmentByTargetCol()).hasEntrySatisfying(
                toCondition(isReference("name", DataTypes.STRING)),
                toCondition(isLiteral("noam")));
        Asserts.assertThat(stmt.outputs()).satisfiesExactly(
            isAlias("foo", isFunction("add", isReference("id"), isLiteral(1L))),
            isAlias("bar", isFunction("subtract")));
    }

    private List<Object[]> execute(Plan plan, Row params) throws Exception {
        TestingRowConsumer consumer = new TestingRowConsumer();
        plan.execute(
            mock(DependencyCarrier.class),
            e.getPlannerContext(clusterService.state()),
            consumer,
            params,
            SubQueryResults.EMPTY
        );
        return consumer.getResult();
    }
}
